/*
**  MKERNEL.g
**
**  purpose : calculate multivariate kernel (density) estimate
**
**  format  : { f,h } = MKERNEL(xs,x,h,w,&kf,mult);
**
**  input   : xs   : (nsxk) matrix, points where density is to be estimated
**            x    : (nxk) matrix, observed data points
**            h    : (kxk) matrix, bandwidth matrix
**            w    : (nx1) vector, weights for different observations
**            &kf  : pointer to procedure containing kernel function
**            mult : 0-1 variable, indicating whether a multiplicative (1) or
**                   spherical kernel is to be used
**
**  output  : f    : (nsx1) vector, estimated density in x
**            h    : (kxk) matrix, bandwidth matrix
**
**  globals :
**
**  remarks : If h<=0, the bandwidth matrix is determined by a rule-of-thumb
**            Resulting bandwidth matrix is proportional to variance-covariance
**            matrix of x
**            Weights are useful for binned data --> w = bincounts
**
**  authors : Dick van Dijk
**
**  date    : First attempt on  7 October  1996
**            Last revision on 17 December 1997
**
*/

proc (2) = mkernel(xs,x,h,w,&kf,mult);
    local ns,k,f,n,invh,deth,sw,i,arg,kff,kfd;
    local kf:proc;

/* Initialization */
ns = rows(xs);
k = cols(xs);
f = zeros(ns,1);
n = rows(x);

/* Determine bandwidth if necessary */
if (h<=0);
    h=real(chol(vcx(x))).*n^(-1/(k+4));
endif;

invh=inv(h);
deth=det(h);

sw=sumc(w);
/* Compute density estimates */
i = 0;
do until (i==ns);
    i=i+1;
    arg = (x-xs[i,.])*invh';
    {kff,kfd} = kf(arg);
    if mult;
        kff=prodc(kff');
    endif;
    f[i] = sumc(kff.*w)./(sw.*deth);
endo;

retp(f,h);
endp;

/*
** This file contains some kernel functions. All functions take an
** n x k matrix u as their argument and return an n x k matrix with 
** the function evaluated in each point of x and an n x k matrix d 
** with the derivative of the function in each point of x.
**
**    k_quart:  quartic kernel function
**    k_epan:   Epanechnikov kernel
**    k_gauss:  Gaussian kernel
**    k_triang: triangular kernel
**    k_rect:   rectangular kernel
**
** usage: {f,d}=k_quart(u);
**
** Extra procedure
**    k_gaussa : computes 4th, 6th and 8th order derivatives of normel kernel
*/

proc (2)=k_quart(u);
 local select;
 select = abs(u).<=1;
 retp( 15/16*((1-u.^2).^2).*select, -15/4*u.*(1-u.^2).*select);
endp;

proc (2)=k_gauss(u);
    retp( pdfn(u), -u.*pdfn(u) );
endp;

proc (2)=k_epan(u);
    local c, s;
    s = abs(u).<1;
    c = 0.75;
    retp( c*(1-u.^2).*s, -0.2*c*u.*s );
endp;

proc (2)=k_epans(u);
    local k,uu,s,c;
    k=cols(u);
    uu=sumc((u.*u)');
    s = abs(uu).<1;
    c=gamma(k/2);
    c = 0.25*(k+2)*k*c/(pi^(k/2));
    retp( c.*(1-uu).*s,0);
endp;

proc (2)=k_rect(u);
    retp( 0.5*(abs(u).<1), 0*u );
endp;

proc (2)=k_trian(u);
    local s, a;
    a = abs(u);
    s = a.<1;
    retp( (1-a).*s, (-(u.>=0) + (u.<=0)).*s );
endp;

proc k_gaussa(u);
    local der4, der6, der8 ;

    der4 = (u.^4-6*u.*2+3).*pdfn(u) ;
    der6 = (u.^6-15*u.^4+45*u.^2-15).*pdfn(u) ;
    der8 = (u.^8-28*u.^6+210*u.^4-420*u.^2+105).*pdfn(u) ;

retp(der4~der6~der8) ;
endp;

