/* Bayesian cointegration analysis in a multivariate Markov trend model
   N(t) = N(t-1) + Gamma(0) + Gamma(1) s(t)
   Y(t) = N(t) + Z(t)
   with Z(t) a VAR(p) model
modified from Richard Paap's "MARKVAR.PRG"
*/


/*
order of variables in "newgdp.txt" [168,7]
Date, Australia, USA, UK, Germany (with dummy for unification), Canada, Japan
*/

new;cls;
library pgraph;
format /rz 8,5;

@ data input @
load yt[168,7]=newgdpD.txt;
yt = yt[.,2:7]; @ strip off date column @
yt=ln(yt)*100;


@ yt has to be a matrix of dimension number of observations times
number of series @

t=rows(yt);
MS = 1;         /*  0: Markov switching off,
                    1: Markov switching on*/
		         

@ output files @
if MS ==1;
txtfile = "bc_res.out";
datfile = "phidraws_bc.dat"; @ stores draws of pi matrix for all possible ranks @
datfile2 = "gam1draws_bc.dat"; @ stores draws of gamma1 matrix for all possible ranks @
pfile = "pdraws_bc.dat";
qfile = "qdraws_bc.dat";
g0file = "g0draws_bc.dat";
n0file = "n0draws_bc.dat";
sigfile = "sigdraws_bc.dat";
phifile = "pidraws_bc.dat";
psifile = "psidraws_bc.dat";
g1file = "g1draws_bc.dat";
else;
txtfile = "var_res.out";
datfile = "phidraws_var.dat";
g0file = "g0draws_var.dat";
n0file = "n0draws_var.dat";
sigfile = "sigdraws_var.dat";
phifile = "pidraws_var.dat";
psifile = "psidraws_var.dat";
endif;

@ simulation settings @
nos = 150000;    @ number of simulations @
noo = 50000;    @ number of burn-in periods @
nob = 1;        @ store every nob-th draw @
stfreq = 10;    @10;@     @ store frequency to harddisk @
lago = 2;       @ lag order @

opt=1;        /* 0: no restrictions of Gamma's,
                1: "business cycle" restrictions drawing from truncated MV normal
		        2: order restrictions via Acceptance/Rejection,
		      */ 



nov = cols(yt); @ number of series@

prior=0;                @ impose prior on Pi yes/no (1/0) @
prtau=0.01;             @ multiplicatieve prior parameter for prior variance Pi @
prmean=zeros(nov,nov);  @ prior mean parameter for Pi @

prh=nov+1;              @ degrees of freedom for Sigma prior @
prmu=1;                 @ multipicative prior parameter for location Sigma prior,
                            see Kleibergen en Paap (2002) @


p00 = 4;	            @ beta prior for transition probabilities -- set all to 1 for uniform prior @
p01 = 1;	            @ NB -- could generalize this to allow different prior for each series... @
q11 = 4;
q10 = 1;

pdf =zeros(nov,1);
MTpdf = zeros(nov,1);

nop = sumc(seqa(1,1,nov)^2);
opslmat = zeros(1,nop);
opslmat2 = zeros(1,nop);

naam1 = "par";naam2 = "gam";

@i=2; do while i<=nop; naam1=naam1|"par"; naam2 = naam2|"gam"; i=i+1; endo;@
create fh=^datfile with naam1,nop,8;
writer(fh,zeros(10,nop));
close(fh);

if MS ==1;
create fh2=^datfile2 with naam2,nop,8;
    writer(fh2,zeros(10,nop));
    close(fh2);
create fh = ^pfile with p,nov,8; call close(fh); psave = {};
create fh = ^qfile with q,nov,8; call close(fh); qsave = {};
create fh = ^g1file with phi,nov*nov,8; call close(fh); g1save = {};
endif;

create fh = ^g0file with g0,nov,8; call close(fh); g0save = {};
create fh = ^n0file with n0,nov,8; call close(fh); n0save = {};
create fh = ^sigfile with sig,nov*(nov+1)/2,8; call close(fh); sigsave = {};
create fh = ^phifile with phi,nov*nov,8; call close(fh); phisave = {};
if lago > 1;
create fh = ^psifile with psi,nov*nov,8; call close(fh); psisave = {};
endif;
tmp={};
tmp2 = {};

xt=ones(t,1)~seqa(0,1,t);
if prior==1;
    prvar=prtau*((yt-xt*(yt/xt))'(yt-xt*(yt/xt)))/t;
else;
    prvar=zeros(nov,nov);
endif;

@ initial starting values Gibbs sampler @
dy=yt[2:t,.]-yt[1:t-1,.];
st=zeros(1,nov)|(dy .lt 0);
ss=st[2:t,.]+st[1:t-1,.];

sumst=zeros(1,nov)|cumsumc(st[2:t,.]);
delst=zeros(1,nov)|st[2:t,.];

p=sumc(ss.==0)/(t-1);
q=sumc(ss.==2)/(t-1);

/* set up storage for state vectors */
if MS ==1;
name = "St"; 
AUstmp = {}; AUstfile = "st_AU_bc.dat"; create fh=^AUstfile with ^name,t,8; call close(fh);
USstmp = {}; USstfile = "st_US_bc.dat"; create fh=^USstfile with ^name,t,8; call close(fh);
UKstmp = {}; UKstfile = "st_UK_bc.dat"; create fh=^UKstfile with ^name,t,8; call close(fh);
GEstmp = {}; GEstfile = "st_GE_bc.dat"; create fh=^GEstfile with ^name,t,8; call close(fh);
JPstmp = {}; CAstfile = "st_CA_bc.dat"; create fh=^CAstfile with ^name,t,8; call close(fh);
CAstmp = {}; JPstfile = "st_JP_bc.dat"; create fh=^JPstfile with ^name,t,8; call close(fh);
endif;

xt=ones(t,1)~seqa(0,1,t)~sumst;
alpha=yt/xt;
res=yt-xt*alpha;
dy=res[2:t,.]-res[1:t-1,.];
if lago>1;
    x1=packr(lagn((ones(1,lago-1).*.dy),((seqa(1,1,lago-1)').*.ones(1,nov))'));
    x1=res[lago:t-1,.]~x1;
else;
    x1=res[lago:t-1,.];
endif;
y1=res[lago+1:t,.]-res[lago:t-1,.];
phi=y1/x1;
res=y1-x1*phi;
res=(yt[1,.]-alpha[1,.])|res;
sig=(res'res)/rows(res);
if lago>1; 
    psi=phi[nov+1:rows(phi),.]; 
    phi=phi[1:nov,.]; 
endif;
delta=0;



@ initalisation @
f=zeros(2,1);
est=zeros(t,nov);
p1=zeros(nov,1); p2=zeros(nov,1); q1=zeros(nov,1); q2=zeros(nov,1); sig1=0; sig2=0;
alpha1=0; alpha2=0; 
phi1=0; phi2=0; psi1=0; psi2=0;
lbp=0; lbp2=0; lbq2=0; lbq=0;
ppq=0; ppq2=0; pgq=0;

counter1 = 0; counter2 = 0; counter3 = 0;

rndseed(1);
i=1;

start = hsec;

do while i<=nos+noo;

   /* x = seqa(1,1,nos+noo); 
    y = zeros(nos+noo,1);
    y[1:i,1] = ones(i,1);
    bar(x,y);
*/
    if MS ==1;
        @ draw states @

        k = 1; 
        do while k<=nov;
            stold=st[.,k];       @store old state @
            repeat:
            res=yt-(ones(t,1)~seqa(0,1,t)~sumst)*alpha;

            @ draw s[t,k] (final state) @
            u0=res[1:t-1,.]|
                (res[t,.]*(1-st[t,k])+(res[t,.]+alpha[k+2,.])*st[t,k]);
            u1=res[1:t-1,.]|
                (res[t,.]*st[t,k]+(res[t,.]-alpha[k+2,.])*(1-st[t,k]));

            if lago>1;
                dres0=u0[2:t,.]-u0[1:t-1,.];
                dres0=packr(lagn((ones(1,lago-1).*.dres0),
                    ((seqa(1,1,lago-1)').*.ones(1,nov))'));
                res0=u0[lago+1:t,.]-u0[lago:t-1,.]*(phi+eye(nov))-dres0*psi;

                dres1=u1[2:t,.]-u1[1:t-1,.];
                dres1=packr(lagn((ones(1,lago-1).*.dres1),
                    ((seqa(1,1,lago-1)').*.ones(1,nov))'));
                res1=u1[lago+1:t,.]-u1[lago:t-1,.]*(phi+eye(nov))-dres1*psi;
            else;
                res0=u0[lago+1:t,.]-u0[lago:t-1,.]*(phi+eye(nov));
                res1=u1[lago+1:t,.]-u1[lago:t-1,.]*(phi+eye(nov));
            endif;

            pdfs=(-0.5*sumc(diag(invpd(sig)*res0'res0)))~
                          (-0.5*sumc(diag(invpd(sig)*res1'res1)));
            pdfs=exp(pdfs-maxc(pdfs')+10);
            f[1]=(1-st[t-1,k])*p[k]+st[t-1,k]*(1-q[k]);
            f[2]=(1-st[t-1,k])*(1-p[k])+st[t-1,k]*q[k];
            f=f.*pdfs';
            f=f/sumc(f);
            st[t,k]=1-(rndu(1,1) .lt f[1]);
            res=u0*(1-st[t,k])+u1*st[t,k];

            @ draw s[t-1,k] till st[lago+1,k] @
            j=t-1;

            do while j>lago;
                u0=res[1:j-1,.]|
                    (res[j,.]*(1-st[j,k])+(res[j,.]+alpha[k+2,.])*st[j,k])|
                    (res[j+1:t,.]*(1-st[j,k])+(res[j+1:t,.]+alpha[k+2,.])*st[j,k]);
                u1=res[1:j-1,.]|
                    (res[j,.].*st[j,k]+(res[j,.]-alpha[k+2,.])*(1-st[j,k]))|
                    (res[j+1:t,.]*st[j,k]+(res[j+1:t,.]-alpha[k+2,.])*(1-st[j,k]));

                if lago>1;
                    dres0=u0[2:t,.]-u0[1:t-1,.];
                    dres0=packr(lagn((ones(1,lago-1).*.dres0),
                        ((seqa(1,1,lago-1)').*.ones(1,nov))'));
                    res0=u0[lago+1:t,.]-u0[lago:t-1,.]*(phi+eye(nov))-dres0*psi;
                    dres1=u1[2:t,.]-u1[1:t-1,.];
                    dres1=packr(lagn((ones(1,lago-1).*.dres1),
                        ((seqa(1,1,lago-1)').*.ones(1,nov))'));
                    res1=u1[lago+1:t,.]-u1[lago:t-1,.]*(phi+eye(nov))-dres1*psi;
                else;
                    res0=u0[lago+1:t,.]-u0[lago:t-1,.]*(phi+eye(nov));
                    res1=u1[lago+1:t,.]-u1[lago:t-1,.]*(phi+eye(nov));
                endif;

                pdfs=(-0.5*sumc(diag(invpd(sig)*res0'res0)))~
                    (-0.5*sumc(diag(invpd(sig)*res1'res1)));
                pdfs=exp(pdfs-maxc(pdfs')+10);
                f[1]=((1-st[j-1,k])*p[k]+st[j-1,k]*(1-q[k]))*((1-st[j+1,k])*p[k]+st[j+1,k]*(1-p[k]));
                f[2]=((1-st[j-1,k])*(1-p[k])+st[j-1,k]*q[k])*((1-st[j+1,k])*(1-q[k])+st[j+1,k]*q[k]);
                f=f.*pdfs';
                f=f/sumc(f);
                st[j,k]=1-(rndu(1,1) .lt f[1]);
                res=u0*(1-st[j,k])+u1*st[j,k];
                j=j-1;
            endo;
    
            @ draw st[1:lago] (first lago states) @

            f[1]=(1-q[k])/(2-p[k]-q[k])*((1-st[lago+1,k])*p[k]+st[lago+1,k]*(1-p[k]));
            f[2]=(1-p[k])/(2-p[k]-q[k])*((1-st[lago+1,k])*(1-q[k])+st[lago+1,k]*q[k]);
            f=f./sumc(f)';
            st[1:lago,k]=1-(rndu(lago,1) .lt f[1]);

            fail1 = (maxc(st[lago+1:t,k])-minc(st[lago+1:t,k])) ==0; @ st = 0 or 1 for all t @
	        if k>1;
                check = 0;
		        jj = 1;
		        do while jj<k;
                    check = check+(sumc(st[.,jj].==st[.,k])==t); @ st[.,k] = st[.,j] for some j @
		            check = check+(sumc(st[.,jj].==(1-st[.,k]))==t); @ st[.,k] = 1-st[.,j] for some j @
		            jj = jj+1;
                endo;
		        fail2 = (check>0);
            else;
		    fail2 = 0;
            endif;	

            if (fail1==1) or (fail2==1);  @ check for ident. states @
                st[.,k]=stold;
                goto repeat;
            endif;
   
            k=k+1;
        endo;

    else;
        st = zeros(t,nov);
    endif;
    sumst=cumsumc(st);@ zeros(1,nov)|cumsumc(st[2:t,.]); @
    delst=st;@ zeros(1,nov)|st[2:t,.]; @
   
    
    
    @ draw p en q @

    ss=st[2:t,.]+st[1:t-1,.];
   
    np=sumc(ss.==0);
    nq=sumc(ss.==2);
    npp=sumc((st[1:t-1,.].==0) .and (st[2:t,.].==1));
    nqq=sumc((st[1:t-1,.].==1) .and (st[2:t,.].==0));
    p=rndbeta(nov,1,np+p00,npp+p01);
    q=rndbeta(nov,1,nq+q11,nqq+q10);

    @ draw sigma @

    res=yt-(ones(t,1)~seqa(0,1,t)~sumst)*alpha;
    if lago>1;
        dres=res[2:t,.]-res[1:t-1,.];
        dres=packr(lagn((ones(1,lago-1).*.dres),
            ((seqa(1,1,lago-1)').*.ones(1,nov))'));
        res=res[lago+1:t,.]-res[lago:t-1,.]*(phi+eye(nov))-dres*psi;
    else;
        res=res[lago+1:t,.]-res[lago:t-1,.]*(phi+eye(nov));
    endif;
    res=(yt[1,.]-alpha[1,.])|res;
    if prior==0;
        chd=chol((res'res));
        sig=chd'rndinvw(nov,rows(res))*chd;
    else;
        chd=chol((res'res+prmu*eye(nov)));
        sig=chd'rndinvw(nov,rows(res)+prh-nov)*chd;
    endif;

    @ draw alpha @
    if MS ==1;
        ydd=yt;
        if lago>1;
            dyt=ydd[2:t,.]-ydd[1:t-1,.];
            dyt=packr(lagn((ones(1,lago-1).*.dyt),
                ((seqa(1,1,lago-1)').*.ones(1,nov))'));
            y1=ydd[lago+1:t,.]-ydd[lago:t-1,.]*(phi+eye(nov))-dyt*psi;
        else;
            y1=ydd[lago+1:t,.]-ydd[lago:t-1,.]*(phi+eye(nov));
        endif;
        cov=chol(invpd(sig));
        ymat=vec(ydd[1,.]*cov')|vec(y1*cov');

        xt=(ones(t,1)~seqa(0,1,t)~sumst);
        x1=(cov.*.xt[lago+1:t,.])-(cov*((phi+eye(nov))').*.xt[lago:t-1,.]);
        tl=1;
        do while tl<lago;
            dxt=xt[lago+1-tl:t-tl,.]-xt[lago-tl:t-1-tl,.];
            x1=x1-(cov*(psi[nov*(tl-1)+1:nov*tl,.]').*.dxt);
            tl=tl+1;
        endo;

        xmat=(cov.*.xt[1,.])|x1;
        alpha=vec(alpha);


        if opt==0;  @ no restrictions on alpha @

            alphat=ymat/xmat; @ alpha is N0, Gamma0, Gamma1 @
            alpvar=invpd(xmat'xmat);
            alpha=chol(alpvar)'rndn((nov+2)*nov,1)+alphat;
            alpha=reshape(alpha,nov,nov+2)';
  
    
        elseif opt==1;  @ restrictions on alpha, enforced by drawing from truncated multivariate normal@
            @ Business cycle restriction -- negative state 1 means @

            alphat=ymat/xmat; @ alpha is N0, Gamma0, Gamma1 @
            alpvar=invpd(xmat'xmat);
            alpha=chol(alpvar)'rndn((nov+2)*nov,1)+alphat; 
            alpha=reshape(alpha,nov,nov+2)';

            /*
            Partition alphat & alpvar as:

            alphat = [mu1|mu2] where mu1 contains all N0 & gamma0 terms, mu2 all gamma1
            alpvar = [v11 ~ v12 |v21 ~ v22] conformable to this
            */

            mu1 = alphat[1:2];
            mu2 = alphat[3:nov+2];
            v11 = alpvar[1:2,1:2];
            v12 = alpvar[1:2,3:nov+2];
            v21 = alpvar[3:nov+2,1:2];
            v22 = alpvar[3:nov+2,3:nov+2];

            kk = 2;
            do while kk<=nov;
	            v11 = v11~alpvar[1:2,(kk-1)*(nov+2)+1:(kk-1)*(nov+2)+2];
	            v12 = v12~alpvar[1:2,(kk-1)*(nov+2)+3:(kk-1)*(nov+2)+nov+2];
	            v21 = v21~alpvar[3:nov+2,(kk-1)*(nov+2)+1:(kk-1)*(nov+2)+2];
	            v22 = v22~alpvar[3:nov+2,(kk-1)*(nov+2)+3:(kk-1)*(nov+2)+nov+2];
                kk = kk+1;
            endo;

            jj = 2;
            do while jj<=nov;
                mu1 = mu1|alphat[(jj-1)*(nov+2)+1:(jj-1)*(nov+2)+2];
                mu2 = mu2|alphat[(jj-1)*(nov+2)+3:(jj-1)*(nov+2)+nov+2];

                tmp11 = alpvar[(jj-1)*(nov+2)+1:(jj-1)*(nov+2)+2,1:2]; 
                tmp22 = alpvar[(jj-1)*(nov+2)+3:(jj-1)*(nov+2)+nov+2,3:nov+2];

                tmp12 = alpvar[(jj-1)*(nov+2)+1:(jj-1)*(nov+2)+2,3:nov+2];
                tmp21 = alpvar[(jj-1)*(nov+2)+3:(jj-1)*(nov+2)+nov+2,1:2];

                kk = 2;
	            do while kk<=nov;
		            tmp11 = tmp11~alpvar[(jj-1)*(nov+2)+1:(jj-1)*(nov+2)+2,(kk-1)*(nov+2)+1:(kk-1)*(nov+2)+2];
		            tmp12 = tmp12~alpvar[(jj-1)*(nov+2)+1:(jj-1)*(nov+2)+2,(kk-1)*(nov+2)+3:(kk-1)*(nov+2)+nov+2];
		            tmp21 = tmp21~alpvar[(jj-1)*(nov+2)+3:(jj-1)*(nov+2)+nov+2,(kk-1)*(nov+2)+1:(kk-1)*(nov+2)+2];
		            tmp22 = tmp22~alpvar[(jj-1)*(nov+2)+3:(jj-1)*(nov+2)+nov+2,(kk-1)*(nov+2)+3:(kk-1)*(nov+2)+nov+2];
		            kk = kk+1;
	            endo;

                v11 = v11|tmp11;
                v12 = v12|tmp12;
                v21 = v21|tmp21;
                v22 = v22|tmp22;

                jj = jj+1;
            endo;

            del21 = v21*invpd(v11);
            mu21 = mu2 - del21*mu1;
            sig221 = v22 - del21*v12;

            cmean = mu21 + del21*vec(alpha[1:2,.]);
       
            @ now draw gamma1 from truncated MVN -- mean=ahat, vcv=avar, diagonal elements right-truncated at -gamma0 @

            M = nov^2;
            MU = cmean; 
            W = sig221;
            WI = invp(sig221);
            C = chol(W)';
            A = mu-100*diag(c);@ones(M,1);@
            B = mu+100*diag(c);@ones(M,1);@
            @ set up restrictions on diagonal elements @
            jj = 1;
            do while jj<=nov;
	            B[(jj-1)*nov+jj] = -alpha[2,jj];
	            jj = jj+1;
            endo;
            R = 100;
            U = rndu(M,R);
            PARM = R;

            {pp,hu,hc} = GHK(M,MU,W,WI,C,A,B,R,U,PARM); 
 
            g1tst = reshape(mu+hc[2:nov^2+1,1],nov,nov)';

            alpha = alpha[1:2,.]|g1tst;

        elseif opt==2;  @ restrictions on alpha via AC/REJ convenient for nov > 2 @

            @ Order restriction -- state 1 means < state 0 means @
	        alphat=ymat/xmat;
	        alpvar=invpd(xmat'xmat);
	        accept = 0; 
            do while accept==0;
	            alpha=chol(alpvar)'rndn((nov+2)*nov,1)+alphat;
	            alpha=reshape(alpha,nov,nov+2)';
		        GG0 = alpha[2,.]';
		        GG1 = alpha[3:nov+2,.]';
		        fail1 = sumc(GG0.<0);
		        fail2 = sumc(diag(GG1).>0);
                if fail1 or fail2;
			        accept = 0;
		        else;
			        accept = 1;
		        endif;
	        endo;
    
        endif;
    else;
        ydd=yt;
        if lago>1;
            dyt=ydd[2:t,.]-ydd[1:t-1,.];
            dyt=packr(lagn((ones(1,lago-1).*.dyt),
                ((seqa(1,1,lago-1)').*.ones(1,nov))'));
            y1=ydd[lago+1:t,.]-ydd[lago:t-1,.]*(phi+eye(nov))-dyt*psi;
        else;
            y1=ydd[lago+1:t,.]-ydd[lago:t-1,.]*(phi+eye(nov));
        endif;
        cov=chol(invpd(sig));
        ymat=vec(ydd[1,.]*cov')|vec(y1*cov');

        xt=(ones(t,1)~seqa(0,1,t));
        x1=(cov.*.xt[lago+1:t,.])-(cov*((phi+eye(nov))').*.xt[lago:t-1,.]);
        tl=1;
        do while tl<lago;
            dxt=xt[lago+1-tl:t-tl,.]-xt[lago-tl:t-1-tl,.];
            x1=x1-(cov*(psi[nov*(tl-1)+1:nov*tl,.]').*.dxt);
            tl=tl+1;
        endo;

        xmat=(cov.*.xt[1,.])|x1;
        alphat=ymat/xmat; @ alpha is N0, Gamma0, Gamma1 @
        alpvar=invpd(xmat'xmat);
        alpha=chol(alpvar)'rndn(2*nov,1)+alphat;
        alpha=reshape(alpha,nov,2)';
        g1tst = zeros(nov,nov);
        alpha = alpha[1:2,.]|g1tst;
    endif;

    @ draw psi @

    if lago>1;
        res=yt-(ones(t,1)~seqa(0,1,t)~sumst)*alpha;
        dres=res[2:t,.]-res[1:t-1,.];
        dres=packr(lagn((ones(1,lago-1).*.dres),
            ((seqa(1,1,lago-1)').*.ones(1,nov))'));
        y1=res[lago+1:t,.]-res[lago:t-1,.]*(phi+eye(nov));
        x1=dres;
        psivar=invpd(x1'x1);
        psihat=psivar*(x1'y1);
        psi=chol(psivar)'rndn((lago-1)*nov,nov)*chol(sig)+psihat;
    endif;

    @ draw phi @

    res=yt-(ones(t,1)~seqa(0,1,t)~sumst)*alpha;
    if lago>1;
        dres=res[2:t,.]-res[1:t-1,.];
        dres=packr(lagn((ones(1,lago-1).*.dres),
            ((seqa(1,1,lago-1)').*.ones(1,nov))'));
        y1=res[lago+1:t,.]-res[lago:t-1,.]-dres*psi;
    else;
        y1=res[lago+1:t,.]-res[lago:t-1,.];
    endif;
    x1=res[lago:t-1,.];

    if prior==0;
        pivar=invpd(x1'x1);
        phihat=y1/x1;
    else;
        pivar=invpd(x1'x1+prvar);
        phihat=pivar*(prvar*prmean+x1'y1);
    endif;
    phi=chol(pivar)'rndn(nov,nov)*chol(sig)+phihat;


    if i>noo and (i%nob==0);     @ store draws of pi for all ranks@

        opslmat=vec(phi)';

       {u,s,v}=svd1(phi);
        rk=nov-1;
        do while rk>0;
            l=inv(powhalf((u[rk+1:nov,rk+1:nov]*u[rk+1:nov,rk+1:nov]')))
                *u[rk+1:nov,rk+1:nov]*s[rk+1:nov,rk+1:nov]*v[rk+1:nov,rk+1:nov]'
                    *inv(powhalf(v[rk+1:nov,rk+1:nov]*v[rk+1:nov,rk+1:nov]'));
            opslmat=opslmat~vec(l)';
            rk=rk-1;
        endo;
        tmp=tmp|opslmat;

        if rows(tmp) .eq stfreq;

            open fh=^datfile for append;
            call writer(fh,tmp);
            call close(fh);
            tmp={};

        endif;

        gamma1 = alpha[3:nov+2,.];
        opsmat2=vec(gamma1)'; @ store draws of gamma1 for all ranks @
        {u,s,v}=svd1(gamma1);
        rk=nov-1;
        do while rk>0;
            l=inv(powhalf((u[rk+1:nov,rk+1:nov]*u[rk+1:nov,rk+1:nov]')))
                *u[rk+1:nov,rk+1:nov]*s[rk+1:nov,rk+1:nov]*v[rk+1:nov,rk+1:nov]'
                    *inv(powhalf(v[rk+1:nov,rk+1:nov]*v[rk+1:nov,rk+1:nov]'));
            opsmat2=opsmat2~vec(l)';
            rk=rk-1;
        endo;
        tmp2=tmp2|opsmat2;
        if MS ==1;
            if rows(tmp2) .eq stfreq;

                open fh=^datfile2 for append;
                call writer(fh,tmp2);
                call close(fh);
                tmp2={};
            endif;

        g1save = g1save|vec(alpha[3:nov+2,.])';
        if rows(g1save) .eq stfreq;
            open fh=^g1file for append;
            call writer(fh,g1save);
            call close(fh);
            g1save={};
        endif;

        endif;

        n0 = alpha[1,.];
	    n0save = n0save|n0;
        if rows(n0save) .eq stfreq;

            open fh=^n0file for append;
            call writer(fh,n0save);
            call close(fh);
            n0save={};
        endif;

        gamma0 = alpha[2,.];
	    g0save = g0save|gamma0;
        if rows(g0save) .eq stfreq;
            open fh=^g0file for append;
            call writer(fh,g0save);
            call close(fh);
        g0save={};
        endif;

	    sigsave = sigsave|vech(sig)';
        if rows(sigsave) .eq stfreq;
            open fh=^sigfile for append;
            call writer(fh,sigsave);
            call close(fh);
            sigsave={};
        endif;

        phisave = phisave|vec(phi)';
        if rows(phisave) .eq stfreq;
            open fh=^phifile for append;
            call writer(fh,phisave);
            call close(fh);
            phisave={};
        endif;
if lago >1;
       psisave = psisave|vec(phi)';
        if rows(psisave) .eq stfreq;
            open fh=^psifile for append;
            call writer(fh,psisave);
            call close(fh);
            psisave={};
        endif;
endif;

        pgq=pgq+(p .gt q);
        p1=p1+p; p2=p2+p^2; q1=q1+q; q2=q2+q^2;

	    

        if MS ==1;
            psave = psave|p';
	        qsave = qsave|q';
            if rows(psave) .eq stfreq;
                open fh=^pfile for append;
                call writer(fh,psave);
                call close(fh);
                psave={};
            endif;

            if rows(qsave) .eq stfreq;
                open fh=^qfile for append;
                call writer(fh,qsave);
                call close(fh);
                qsave={};
            endif;
       

	        AUstmp = AUstmp|st[.,1]';
	        USstmp = USstmp|st[.,2]';
	        UKstmp = UKstmp|st[.,3]';
	        GEstmp = GEstmp|st[.,4]';
	        CAstmp = CAstmp|st[.,5]';
	        JPstmp = JPstmp|st[.,6]';

            if rows(AUstmp) .eq stfreq;
                open fh=^AUstfile for append;
                call writer(fh,AUstmp);
                call close(fh);
                AUstmp={};
            endif;

            if rows(USstmp) .eq stfreq;
                open fh=^USstfile for append;
                call writer(fh,USstmp);
                call close(fh);
                USstmp={};
            endif;

            if rows(UKstmp) .eq stfreq;
                open fh=^UKstfile for append;
                call writer(fh,UKstmp);
                call close(fh);
                UKstmp={};
            endif;

            if rows(GEstmp) .eq stfreq;
                open fh=^GEstfile for append;
                call writer(fh,GEstmp);
                call close(fh);
                GEstmp={};
            endif;

            if rows(CAstmp) .eq stfreq;
                open fh=^CAstfile for append;
                call writer(fh,CAstmp);
                call close(fh);
                CAstmp={};
            endif;

            if rows(JPstmp) .eq stfreq;
                open fh=^JPstfile for append;
                call writer(fh,JPstmp);
                call close(fh);
                JPstmp={};
            endif;
        endif;

        sig1=sig1+sig; sig2=sig2+sig^2;
        alpha1=alpha1+alpha; alpha2=alpha2+alpha^2;
        phi1=phi1+phi; phi2=phi2+phi^2;
        if lago>1;
            psi1=psi1+psi; psi2=psi2+psi^2;
        endif;
        est=est+st;
        ppq=ppq+p+q;
        ppq2=ppq2+(p+q)^2;
        lbp=lbp+1/(1-p); lbp2=lbp2+(1/(1-p))^2;
        lbq=lbq+1/(1-q); lbq2=lbq2+(1/(1-q))^2;
     endif;
     


    i=i+1;
endo;

finish = hsec;

/*if rows(tmp)>0;
    open fh=^datfile for append;
    call writer(fh,tmp);
    call close(fh);
endif;
*/

if prior ==1; @ compute prior heights @
    pdfpr=priorhgt(prh,prmu*eye(nov),prmean,prvar,nos);
else;
    pdfpr={};
    rk=nov; do while rk>0; pdfpr=pdfpr|((1/sqrt(2*pi))^(rk*rk)); rk=rk-1; endo;
endif;

nos=nos/nob;     @ compute number of valid draws @
est=est/nos;     @ expected state @
xy(seqa(1,1,t),est);

@ compute Bayes factors for cointegration rank @

pdf=pdf/nos;
pdfpo={};
open fh=^datfile for read;
dr=readr(fh,10);
dr=readr(fh,nos);
close(fh);
rk=nov;
do while rk>0;
    vdr=dr[.,1:rk*rk];
    {a,b}=mkernel(zeros(1,rk*rk),vdr,-1,ones(nos,1),&k_gauss,1);
    pdfpo=pdfpo|a;
    if rk>1;  dr=trimr(dr',rk*rk,0)'; endif;
    rk=rk-1;
endo;

    @ compute Bayes factors for common Markov trends @

    MTpdf=MTpdf/nos;

    pdfpoMT={};
if MS ==1;
    open fh=^datfile2 for read;
    dr=readr(fh,10);
    dr=readr(fh,nos);
    close(fh);

    rk=nov;
    do while rk>0;
        vdr=dr[.,1:rk*rk];
        {a,b}=mkernel(zeros(1,rk*rk),vdr,-1,ones(nos,1),&k_gauss,1);
        pdfpoMT=pdfpoMT|a;
        if rk>1;  dr=trimr(dr',rk*rk,0)'; endif;
        rk=rk-1;
    endo;
endif;
output file=^txtfile reset;

kk=(ln(pdfpo)-ln(pdfpr))|0;
kk = missrv(kk.*(kk./=__INFn),-500);
kk = missrv(kk.*(kk./=__INFp),500); 

@ check for +/- INF @
 

kkMT = (ln(pdfpoMT)-ln(pdfpr))|0;
kkMT = missrv(kkMT.*(kkMT./=__INFn),-500);
kkMT = missrv(kkMT.*(kkMT./=__INFp),500);   


"==========================================================";
"date" (date)';
"number of simulations:" nos;
"number of burning-in periods:" noo;
"time = " (finish-start)/100 "seconds";
"";
"----------------------------------------------------------";
"Inference on Cointegration Rank:"
"log Bayes factors"
seqa(0,1,nov+1)~kk;
"posterior probabilities"
seqa(0,1,nov+1)~exp(kk)./sumc(exp(kk))';
if MS ==1;
"----------------------------------------------------------";
"Inference on Common Markov trends:"
"log Bayes factors"
seqa(0,1,nov+1)~kkMT;
"posterior probabilities"
seqa(0,1,nov+1)~exp(kkMT)./sumc(exp(kkMT))';
"----------------------------------------------------------";
"means + variances of p and q";
(p1/nos)~sqrt(p2/nos-(p1/nos)^2);
(q1/nos)~sqrt(q2/nos-(q1/nos)^2);
endif;
"means + variances of N0, Gamma0 and Gamma1 "
alpha1/nos~sqrt(alpha2/nos-(alpha1/nos)^2);
"means + variances of pi"
phi1/nos~sqrt(phi2/nos-(phi1/nos)^2);
if lago>1;
    "means + variances of psi"
    psi1/nos~sqrt(psi2/nos-(psi1/nos)^2);
endif;
"means + variances of sigma"
sig1/nos~sqrt(sig2/nos-(sig1/nos)^2);
if MS==1;
"means + variances of duration of recession and expansion";
(lbp/nos)~sqrt(lbp2/nos-(lbp/nos)^2);
(lbq/nos)~sqrt(lbq2/nos-(lbq/nos)^2);
"Probability p > q";
pgq/nos;
endif;
if prior==1;
    "------------------------------------------------------------";
    "prior parameters sigma (degrees of freedom)";
    prh;
    "and (scale)"
    prmu*eye(nov);
    "prior parameters Pi (mean)"
    prmean;
    "and (tau)";
    prtau;
else;
    "uniformative prior";
endif;
"==========================================================";
output off;
@system;@

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  Q. Gibbs's Sampler Simulator (GSS)
     The GSS simulates HC, but does not simulate P or HU.  It assumes
that PARM[1,1] gives the number of rounds to be used in the construction
of the random draws (JL). It assumes that U is an MxJL array
of uniform [0,1] random numbers.  It returns only HC.

++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

PROC 3 = GSS(M,MU,W,WI,C,A,B,R,U,PARM);
LOCAL HC,JL,I,J,K,MK,TA,TB,E,F,G,D,S,HT,V,P,HU,UI,UIKJ;
local wimkk,wmkmki,wmkk,vbar,vvpbar,condmukf,condsdk,csdk,cmuk,seedgss;
local timelim, vchk;

@vah: guard against possible division by 0, by adding 1.e-100 to denominators@

ui=u;

condmukf=zeros(m,m-1);
condsdk =zeros(m,1);
k=1;
do until k>m;
      IF K == 1;                             @INDICES EXCLUDING K@
        MK = SEQA(2,1,M-1);
        ELSE;
         IF K == M;
         MK = SEQA(1,1,M-1);
         ELSE;
         MK = SEQA(1,1,K-1)|SEQA(K+1,1,M-K);
         ENDIF;
       ENDIF;
      wimkk  = WI[MK,K];
      wmkk   = W[MK,K];
      wmkmki =
   wi[mk,mk] - (wimkk*wimkk')/(wi[k,k]+1.e-100); @INVERSE OF W EXCL.ROW/COL K@
      condsdk[k,1] = sqrt(w[k,k] - (wmkk')*wmkmki*wmkk);  @CONDIT. SD OF V[K]@
      condmukf[k,.] = (wmkk')*wmkmki;  @FACTOR FOR CONDIT. MEAN OF V[K]-MU[K]@
k=k+1;
endo;
@condmukf~condsdk;@

/*begin*/
vbar=0;
vvpbar=0;
JL = PARM[1,1];                              @NUMBER OF GIBBS ROUNDS@
I = 0;
V = (B+A)/2-MU;                              @INITIAL VECTOR@
/*timelim=hsec1990();*/
DO WHILE I < R;                              @LOOP OVER REPETITIONS@
  I = I+1;
  if i > 1;
      seedgss=ui[1,1]*1.e+9;
      ui=rndus(m,jl,seedgss);
  endif;
  J = 0;
  DO WHILE J < JL;                           @DRAW UNTIL LIMIT@
    J = J+1;
    K = 0;                                   @START RECURSIVE STEPS@
    DO WHILE K < M;                          @STEP OVER K@
      K = K+1;
      IF K == 1;                             @INDICES EXCLUDING K@
        MK = SEQA(2,1,M-1);
        ELSE;
         IF K == M;
         MK = SEQA(1,1,M-1);
         ELSE;
         MK = SEQA(1,1,K-1)|SEQA(K+1,1,M-K);
         ENDIF;
       ENDIF;
      csdk=condsdk[k,1];
      cmuk=condmukf[k,.]*v[mk,1]; @cmuk~csdk;@
      TA = CDFN((A[K]-MU[K]-cmuk)/(csdk+1.e-100)); @BOUNDS@
      TB = CDFN((B[K]-MU[K]-cmuk)/(csdk+1.e-100));
      UIKJ = UI[K,J];
@"ta tb uikj k"; ta tb uikj k;@
      V[K] = cmuk + csdk*CDFINVN(UIKJ*TA+(1-UIKJ)*TB); @DRAW FROM TRUNCATED N@
    ENDO;
  ENDO;

@vchk;@
  vbar=vbar+v;
  vvpbar=vvpbar+v*v';
/*  if (hsec1990()-timelim) gt hseclim;*/
/*      @give up@*/
/*      r=j;*/
/*      j=r;*/
/*  endif;*/
ENDO;
vbar=vbar/r;
vvpbar=(vvpbar/r - w)*wi/2;
hc=(1|vbar)~(vbar'|vvpbar);
P = -999;
HU = -999*ones(M+1,M+1);
RETP(P,HU,vbar);
ENDP;

/*
** CDFINVN --   computes the inverse normal cdf. The algorithm used is taken
**              from Kennedy and Gentle, STATISTICAL COMPUTING, p. 95.
**
** vah 22:57:35.78, Sun, Jul 23, 1989
** Taken intact from V149b MAXPROC3.ARC
**
**              xp=cdfinvn(p);
**
**             INPUT
**              p -- nx1 vector of probabilities (that is, 0 < p < 1)
**             OUTPUT
**             xp -- the corresponding vector of values such that: cdfn(xp)=p;
**
**              This algorithm seems to be very accurate.  It is
**              rated to be accurate to 1e-7.  However, it appears
**              to be substantially more accurate -- approximately
**              1e-10.
**
*/

proc 1=cdfinvn(p);
local lim, p0, p1, p2, p3, p4, q0, q1, q2, q3, q4, maskgt, maskeq, sgn,
y, xp, pn;

@ Constants @

lim =  1e-20;

p0  = - 0.322232431088;             q0  =   0.0993484626060;
p1  = - 1.0;                        q1  =   0.588581570495;
p2  = - 0.342242088547;             q2  =   0.531103462366;
p3  = - 0.0204231210245;            q3  =   0.103537752850;
p4  = - 0.453642210148*1e-4;        q4  =   0.38560700634*1e-2;


@ Main body of code @

 @ Create masks for handling p > .5 and p >= .5 @

maskgt=(p .> 0.5);
maskeq=(p .ne 0.5);
sgn=missrv(miss(maskgt,0),-1);

 @ Convert p > .5 --> 1-p @

pn=(maskgt-p).*sgn;   clear maskgt;

 @ Computation of function for p < 0.5 @

y=sqrt(-2*ln(pn));     clear pn;
xp=y + ((((y*p4 + p3).*y + p2).*y + p1).*y + p0)./
       ((((y*q4 + q3).*y + q2).*y + q1).*y + q0); clear y;

 @ Convert results for p > .5 and p = .5 @

xp=(xp.*sgn).*maskeq;
clear sgn, maskeq;
retp(xp);
endp;

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  M. Geweke-Hajivassiliou-Keane Simulator (GHK)
     The GHK procedure assumes that U is an array of uniform [0,1] variates,
independent within columns, and in general independent between columns.
It does not use PARM.  It returns both HU and HC. 

++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

PROC 3 = GHK(M,MU,W,WI,C,A,B,R,U,PARM);
LOCAL J,II,TA,TB,TT,WGT,V,HU,P,HC;
@vah: guard against possible division by 0, by adding 1.e-100 to denominators@
J = 1;
II = 1;
TA = CDFN((A[1,1]-MU[1,1])/(C[1,1]+1.e-100))*ONES(1,R); @1ST COMPONENT BOUNDS@
TB = CDFN((B[1,1]-MU[1,1])/(C[1,1]+1.e-100))*ONES(1,R);
/*TT = CDFNI(U[1,.].*TA+(1-U[1,.]).*TB);*/
TT = CDFINVN(U[1,.].*TA+(1-U[1,.]).*TB);
WGT = TB-TA;
DO WHILE J < M;                                         @LOOP OVER COMPONENTS@
  J = J+1;
  @COMPONENT J BOUNDS@
  TA = CDFN(((A[J,1]-MU[J,1])*ONES(1,R)-C[J,II]*TT)/(C[J,J]+1.e-100));
  TB = CDFN(((B[J,1]-MU[J,1])*ONES(1,R)-C[J,II]*TT)/(C[J,J]+1.e-100));
/*  TT = TT|CDFNI(U[J,.].*TA+(1-U[J,.]).*TB);           @COMPONENT J DRAW@*/
  TT = TT|CDFINVN(U[J,.].*TA+(1-U[J,.]).*TB);           @COMPONENT J DRAW@
  II = II|J;
  WGT = WGT.*(TB-TA);                                   @ ACCUMULATE WEIGHT@
ENDO;
V = C*TT;                                               @TRANSFORM TO V-MU@
TT = (ONES(M,1)*WGT).*V;
P = SUMC(WGT')/R;
HU = WI*V*(TT')*WI/R-P*WI;
V = SUMC(TT')/R; @WI*SUMC(TT')/R; @
HU = (P|V)~((V')|(HU/2));
IF P > 0;
  HC = HU/P;
  ELSE;
  HC = -999*ONES(M+1,M+1);
ENDIF;
RETP(P,HU,HC);
ENDP;   
